import time

import requests

from morphapi.utils.data_io import connected_to_internet

mouselight_base_url = "http://ml-neuronbrowser.janelia.org/"


def request(url, verify=True):
    """
    Sends a request to a url

    :param url:

    """
    if not connected_to_internet():
        raise ConnectionError(
            "You need to have an internet connection to send requests."
        )

    try:
        _DEFAULT_CIPHERS = requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS
        if not verify:
            requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS = (
                _DEFAULT_CIPHERS + ":HIGH:!DH:!aNULL"
            )
        response = requests.get(url, verify=verify)
    finally:
        requests.packages.urllib3.util.ssl_.DEFAULT_CIPHERS = _DEFAULT_CIPHERS

    if response.ok:
        return response
    else:
        exception_string = "URL request failed: {}".format(response.reason)
    raise ValueError(exception_string + f" ; url: {url}")


def query_mouselight(query):
    """
    Sends a GET request, not currently used for anything.

    :param query:

    """
    if not connected_to_internet():
        raise ConnectionError(
            "You need an internet connection for API queries, sorry."
        )

    base_url = "http://ml-neuronbrowser.janelia.org/"

    full_query = base_url + query

    # send the query, package the return argument as a json tree
    response = requests.get(full_query)
    if response.ok:
        json_tree = response.json()
        if json_tree["success"]:
            return json_tree
        else:
            exception_string = "did not complete api query successfully"
    else:
        exception_string = "API failure. Allen says: {}".format(
            response.reason
        )

    # raise an exception if the API request failed
    raise ValueError(exception_string)


def post_mouselight(url, query=None, clean=False, attempts=3):
    """
    sends a POST request to a user URL. Query can be either a string (in which case clean should be False) or a dictionary.

    :param url:
    :param query: string or dictionary with query   (Default value = None)
    :param clean: if not clean, the query is assumed to be in JSON format (Default value = False)
    :param attempts: number of attempts  (Default value = 3)

    """
    if not connected_to_internet():
        raise ConnectionError(
            "You need an internet connection for API queries, sorry."
        )

    request = None
    if query is not None:
        for i in range(attempts):
            try:
                if not clean:
                    time.sleep(0.01)  # avoid getting an error from server
                    request = requests.post(url, json={"query": query})
                else:
                    time.sleep(0.01)  # avoid getting an error from server
                    request = requests.post(url, json=query)
            except Exception as e:
                exception = e
                request = None
                print(
                    "MouseLight API query failed. Attempt {} of {}".format(
                        i + 1, attempts
                    )
                )
            if request is not None:
                break

        if request is None:
            raise ConnectionError(
                "\n\nMouseLight API query failed with error message:\n{}.\
                        \nPerhaps the server is down, visit 'http://ml-neuronbrowser.janelia.org' to find out.".format(
                    exception
                )
            )
    else:
        raise NotImplementedError

    if request.status_code == 200:
        jreq = request.json()
        if "data" in list(jreq.keys()):
            return jreq["data"]
        else:
            return jreq
    else:
        raise Exception(
            "Query failed to run by returning code of {}. {} -- \n\n{}".format(
                request.status_code, query, request.text
            )
        )



from morphapi.api.allenmorphology import AllenMorphology
from morphapi.api.mouselight import MouseLightAPI
from morphapi.api.neuromorphorg import NeuroMorpOrgAPI


def test_neuromorpho_download():
    api = NeuroMorpOrgAPI()
    metadata, _ = api.get_neurons_metadata(
        size=2,  # Can get the metadata for up to 500 neurons at the time
        species="mouse",
        cell_type="pyramidal",
        brain_region="neocortex",
    )

    if len(metadata) != 2:
        raise ValueError("Incorrect metadata length")

    neurons = api.download_neurons(metadata)

    if len(neurons) != len(metadata):
        raise ValueError

    neurons = [neuron.create_mesh()[1] for neuron in neurons]

    # Test no load
    assert api.download_neurons(metadata, load_neurons=False)[0].points is None

    # Test failure
    metadata[0]["neuron_id"] = "BAD ID"
    metadata[0]["neuron_name"] = "BAD NAME"
    neurons = api.download_neurons([metadata[0]])

    assert neurons[0].data_file.name == "BAD ID.swc"
    assert neurons[0].points is None


def test_mouselight_download():
    mlapi = MouseLightAPI()

    neurons_metadata = mlapi.fetch_neurons_metadata(
        filterby="soma", filter_regions=["MOs"]
    )
    neurons_metadata = sorted(neurons_metadata, key=lambda x: x["idString"])

    neurons = mlapi.download_neurons(neurons_metadata[0])

    neurons = [neuron.create_mesh()[1] for neuron in neurons]

    # Test no load
    assert (
        mlapi.download_neurons(neurons_metadata[0], load_neurons=False)[
            0
        ].points
        is None
    )

    # Test failure
    neurons_metadata[0]["idString"] = "BAD ID"
    neurons = mlapi.download_neurons(neurons_metadata[0])

    assert neurons[0].data_file.name == "BAD ID.swc"
    assert neurons[0].points is None

    # Test filter without atlas
    filtered_neurons_metadata = mlapi.filter_neurons_metadata(
        neurons_metadata, filterby="soma", filter_regions=["MOs6b"]
    )
    assert all(
        [i["brainArea_acronym"] == "MOs6b" for i in filtered_neurons_metadata]
    )

    # Test filter with atlas
    atlas = mlapi.fetch_default_atlas()

    filtered_neurons_metadata_atlas = mlapi.filter_neurons_metadata(
        neurons_metadata,
        filterby="soma",
        filter_regions=["MOs6b"],
        atlas=atlas,
    )
    assert all(
        [
            i["brainArea_acronym"] == "MOs6b"
            for i in filtered_neurons_metadata_atlas
        ]
    )
    assert filtered_neurons_metadata == filtered_neurons_metadata_atlas


def test_allen_morphology_download():
    am = AllenMorphology()

    # Select some mouse neurons in the primary visual cortex
    neurons_df = am.neurons.loc[
        (am.neurons.species == "Mus musculus")
        & (am.neurons.structure_area_abbrev == "VISp")
    ].loc[[2, 120, 505]]

    # Download some neurons
    neurons = am.download_neurons(neurons_df["id"].values)

    neurons = [neuron.create_mesh()[1] for neuron in neurons]

    # Test no load
    assert all(
        i.points is None
        for i in am.download_neurons(
            neurons_df["id"].values, load_neurons=False
        )
    )

    # Test failure
    neurons_df.loc[2, "id"] = "BAD ID"
    neurons = am.download_neurons(neurons_df["id"].values)

    assert neurons[0].data_file.name == "BAD ID.swc"
    assert neurons[0].points is None



from random import choice

import numpy as np
import pytest
from vedo import Mesh

from morphapi.morphology.morphology import Neuron
from morphapi.utils.data_io import listdir


@pytest.fixture
def neuron():
    files = listdir("tests/data")
    return Neuron(data_file=choice(files))


args = [
    (3, "salmon", "darkseagreen", "orangered", "blackboard", "blue", True),
    (3, "salmon", None, None, "blackboard", "blue", False),
    (3, "salmon", None, None, None, None, False),
    (5, "salmon", "darkseagreen", "orangered", "blackboard", "blue", False),
]


@pytest.mark.parametrize("radius,soma,apical,basal,axon,whole,cache", args)
def test_create_mesh(neuron, radius, soma, apical, basal, axon, whole, cache):
    components, neuron = neuron.create_mesh(
        neurite_radius=radius,  #
        soma_color=soma,  # Specify colors [see vedo.colors for more details]
        apical_dendrites_color=apical,
        basal_dendrites_color=basal,
        axon_color=axon,
        whole_neuron_color=whole,
        use_cache=cache,
    )

    if not isinstance(neuron, Mesh):
        raise ValueError

    if not isinstance(components, dict):
        raise ValueError

    for ntp in Neuron._neurite_types:
        if ntp not in components.keys():
            raise ValueError
        if components[ntp] is not None:
            if not isinstance(components[ntp], Mesh):
                raise ValueError


def test_create_mesh_args(neuron):
    neuron.create_mesh(
        neurite_radius=np.random.uniform(2, 20), neuron_color="red"
    )
    neuron.create_mesh(neuron_number=1, cmap="Reds")


def test_empty_neuron(caplog):
    file = sorted(listdir("tests/data"))[0]
    neuron = Neuron(file, load_file=False)

    assert str(neuron.data_file) == "tests/data/example1.swc"

    caplog.clear()
    neuron.create_mesh()

    assert caplog.messages == [
        "No data loaded, you can use the 'load_from_file' method to try to load the file."
    ]

    neuron.load_from_file()

    caplog.clear()
    neuron.create_mesh()
    assert caplog.messages == []



import os

import requests
import yaml


def listdir(fld):
    """
    List the files into a folder with the coplete file path instead of the relative file path like os.listdir.

    :param fld: string, folder path

    """
    if not os.path.isdir(fld):
        raise FileNotFoundError("Could not find directory: {}".format(fld))

    return [os.path.join(fld, f) for f in os.listdir(fld)]


def get_file_name(filepath):
    # Returns just the name, no complete path or extension
    return os.path.splitext(os.path.basename(filepath))[0]


# ------------------------------ Load/Save data ------------------------------ #


def save_yaml(filepath, content, append=False, topcomment=None):
    """
    Saves content to a yaml file

    :param filepath: path to a file (must include .yaml)
    :param content: dictionary of stuff to save

    """
    if not filepath.endswith(".yaml") and not filepath.endswith(".yml"):
        raise ValueError(
            f"filepath is invalid {filepath}. Should end with yaml or yml"
        )

    if not append:
        method = "w"
    else:
        method = "w+"

    with open(filepath, method) as yaml_file:
        if topcomment is not None:
            yaml_file.write(topcomment)
        yaml.dump(content, yaml_file, default_flow_style=False, indent=4)


def load_yaml(filepath):
    """
    Load a YAML file

    :param filepath: path to yaml file

    """
    if filepath is None or not os.path.isfile(filepath):
        raise ValueError("unrecognized file path: {}".format(filepath))
    if "yml" not in filepath and "yaml" not in filepath:
        raise ValueError("unrecognized file path: {}".format(filepath))
    return yaml.load(open(filepath), Loader=yaml.FullLoader)


# ----------------------------- Internet queries ----------------------------- #
def connected_to_internet(url="http://www.google.com/", timeout=5):
    """
    Check that there is an internet connection

    :param url: url to use for testing (Default value = 'http://www.google.com/')
    :param timeout:  timeout to wait for [in seconds] (Default value = 5)
    """

    try:
        _ = requests.get(url, timeout=timeout)
        return True
    except requests.ConnectionError:
        print("No internet connection available.")
    return False


# ---------------------------------------------------------------------------- #
#                               Data manipulation                              #
# ---------------------------------------------------------------------------- #
def flatten_list(lst):
    """
    Flattens a list of lists

    :param lst: list

    """
    flatten = []
    for item in lst:
        if isinstance(item, list):
            flatten.extend(item)
        else:
            flatten.append(item)
    return flatten


def is_any_item_in_list(L1, L2):
    """
    Checks if an item in a list is in another  list

    :param L1:
    :param L2:

    """
    # checks if any item of L1 is also in L2 and returns false otherwise
    inboth = [i for i in L1 if i in L2]
    if inboth:
        return True
    else:
        return False



from vedo import Plotter

from morphapi.api.neuromorphorg import NeuroMorpOrgAPI

api = NeuroMorpOrgAPI()

# ---------------------------- Downloading metadata --------------------------- #
# Get metadata for pyramidal neurons from the mouse cortex.
metadata, _ = api.get_neurons_metadata(
    size=10,  # Can get the metadata for up to 500 neurons at the time
    species="mouse",
    cell_type="pyramidal",
    brain_region="neocortex",
)

# To get a list of available query fields: print(api.fields)
# To get a list of valid values for a field: print(api.get_fields_values(field))

print("Neurons metadata:")
print(metadata[0])

# ---------------------------- Download morphology --------------------------- #
neurons = api.download_neurons(metadata[5])


# ------------------------------- Visualisation ------------------------------ #
print("creating meshes")
neurons = [neuron.create_mesh()[1] for neuron in neurons]

print("visualizing")
vp = Plotter(shape=(1, len(neurons)), axes=1)

vp.show(neurons)



from vedo import Plotter

from morphapi.api.mpin_celldb import MpinMorphologyAPI

api = MpinMorphologyAPI()

# ----------------------------- Download dataset ----------------------------- #
"""
    If it's the first time using this API, you'll have to download the dataset
    with all of the neurons' data.
"""
api.download_dataset()


# You can then inspect metadata about all neurons:
print(api.neurons_df.head())

# and load a few neurons
neurons = api.load_neurons(list(api.neurons_df.index[:10]))

# ------------------------------- Visualisation ------------------------------ #
print("creating meshes")
neurons = [neuron.create_mesh()[1] for neuron in neurons]

print("visualizing")
vp = Plotter(shape=(1, len(neurons)), axes=1)

vp.show(neurons)



import logging
import os

import numpy as np
import pandas as pd

try:
    from allensdk.core.cell_types_cache import CellTypesCache
except ModuleNotFoundError:
    raise ModuleNotFoundError(
        'You need to install the allen sdk package to use AllenMorphology:  "pip install allensdk"'
    )

from morphapi.morphology.morphology import Neuron
from morphapi.paths_manager import Paths
from morphapi.utils.data_io import connected_to_internet

logger = logging.getLogger(__name__)


class AllenMorphology(Paths):
    """Handles the download of neuronal morphology data from the Allen database."""

    def __init__(self, *args, **kwargs):
        """
        Initialise API interaction and fetch metadata of neurons in the Allen Database.
        """
        if not connected_to_internet():
            raise ConnectionError(
                "You will need to be connected to the internet to use the AllenMorphology class to download neurons"
            )

        Paths.__init__(self, *args, **kwargs)

        # Create a Cache for the Cell Types Cache API
        self.ctc = CellTypesCache(
            manifest_file=os.path.join(
                self.allen_morphology_cache, "manifest.json"
            )
        )

        # Get a list of cell metadata for neurons with reconstructions, download if necessary
        self.neurons = pd.DataFrame(
            self.ctc.get_cells(require_reconstruction=True)
        )
        self.n_neurons = len(self.neurons)

        if not self.n_neurons:
            raise ValueError(
                "Something went wrong and couldn't get neurons metadata from Allen"
            )

        self.downloaded_neurons = self.get_downloaded_neurons()

    def get_downloaded_neurons(self):
        """
        Get's the path to files of downloaded neurons
        """
        return [
            os.path.join(self.allen_morphology_cache, f)
            for f in os.listdir(self.allen_morphology_cache)
            if ".swc" in f
        ]

    def build_filepath(self, neuron_id):
        """
        Build a filepath from neuron's metadata.
        """
        return os.path.join(
            self.allen_morphology_cache, "{}.swc".format(neuron_id)
        )

    def download_neurons(self, ids, load_neurons=True, **kwargs):
        """
            Download neurons and return neuron reconstructions (instances
            of Neuron class)

        :param ids: list of integers with neurons IDs

        """
        if isinstance(ids, np.ndarray):
            ids = ids.tolist()
        if not isinstance(ids, (list)):
            ids = [ids]

        neurons = []
        for neuron_id in ids:
            neuron_file = self.build_filepath(neuron_id)
            load_current_neuron = load_neurons
            logger.debug(
                "Downloading neuron '%s' to %s", neuron_id, neuron_file
            )

            # Download file
            try:
                self.ctc.get_reconstruction(neuron_id, file_name=neuron_file)
            except Exception as exc:
                logger.error(
                    "Could not fetch the neuron %s for the following reason: %s",
                    neuron_id,
                    str(exc),
                )
                load_current_neuron = False

            # Reconstruct neuron
            neurons.append(
                Neuron(
                    neuron_file,
                    neuron_name=str(neuron_id),
                    load_file=load_current_neuron,
                    **kwargs,
                )
            )

        return neurons



import logging
import os
from functools import partial

from morphapi.morphology.morphology import Neuron
from morphapi.paths_manager import Paths
from morphapi.utils.webqueries import connected_to_internet
from morphapi.utils.webqueries import request

logger = logging.getLogger(__name__)
request_no_ssl = partial(request, verify=False)


class NeuroMorpOrgAPI(Paths):
    _base_url = "http://neuromorpho.org/api/neuron"

    _version = "CNG version"  # which swc version, standardized or original

    def __init__(self, *args, **kwargs):
        if not connected_to_internet():
            raise ConnectionError(
                "You will need to be connected to the internet to use the NeuroMorpOrgAPI class to download neurons"
            )

        Paths.__init__(self, *args, **kwargs)

        # Check that neuromorpho.org is not down
        try:
            request_no_ssl("http://neuromorpho.org/api/health")
        except Exception as e:
            raise ConnectionError(
                f"It seems that neuromorphos API is down: {e}"
            )

        # Fields contains the types of fields that can be used to restrict queries
        self.fields = request_no_ssl(self._base_url + "/fields").json()[
            "Neuron Fields"
        ]

    def get_fields_values(self, field):
        """
        Returns the list of allowed values for a given query field
        """
        return list(
            request_no_ssl(self._base_url + f"/fields/{field}").json()[
                "fields"
            ]
        )

    def get_neurons_metadata(self, size=100, page=0, **criteria):
        """
        Uses the neuromorpho API to download metadata about neurons.
        Criteri can be used to restrict the search to neurons of interest/
        http://neuromorpho.org/apiReference.html

        Neuromorpho.org  paginates it's requests so not all neurons metadata
        can be returned at once

        :param size: int in range [0, 500]. Number of neurons whose metadata can be returned at the same time
        :param page: int > 0. Page number. The number of pages depends on size and on how many neurons match the criteria
        :param criteria: use keywords to restrict the query to neurons that match given criteria.
                keywords should be pass as "field=value". Then only neuron's whose 'field'
                attribute has value 'value' will be returned.
        """

        if size < 0 or size > 500:
            raise ValueError(
                f"Invalid size argument: {size}. Size should be an integer between 0 and 500"
            )
        if page < 0:
            raise ValueError(
                f"Invalid page argument: {page}. Page should be an integer >= 0"
            )

        url = self._base_url + "/select?q="

        for n, (crit, val) in enumerate(criteria.items()):
            if crit not in self.fields:
                raise ValueError(
                    f"Query criteria {crit} not in available fields: {self.fields}"
                )
            elif val not in self.get_fields_values(crit):
                raise ValueError(
                    f"Query criteria value {val} for field {crit} not valid."
                    + f"Valid values include: {self.get_fields_values(crit)}"
                )

            if isinstance(val, list):
                raise NotImplementedError("Need to join the list")

            if n > 0:
                url += "&fq="
            url += f"{crit}:{val}"

        url += f"&size={int(size)}&page={int(page)}"
        req = request_no_ssl(url)

        if not req.ok:
            raise ValueError(f"Invalid query with url: {url}")
        else:
            neurons = req.json()

        page = neurons["page"]
        neurons = neurons["_embedded"]["neuronResources"]

        logger.info(
            f"Found metadata for {page['totalElements']} neurons [{page['totalPages']} pages in total]. "
            f"Returning metadata about {len(neurons)} neurons from page {page['number']}"
        )

        return neurons, page

    def get_neuron_by_id(self, nid):
        """
        Get a neuron's metadata given it's id number
        """
        return request_no_ssl(self._base_url + f"/id/{nid}").json()

    def get_neuron_by_name(self, nname):
        """
        Get a neuron's metadata given it's name
        """
        return request_no_ssl(self._base_url + f"/name/{nname}").json()

    def build_filepath(self, neuron_id):
        """
        Build a filepath from a neuron ID.
        """
        return os.path.join(self.neuromorphorg_cache, f"{neuron_id}.swc")

    def download_neurons(
        self, neurons, _name=None, load_neurons=True, **kwargs
    ):
        """
        Downloads neuronal morphological data and saves it to .swc files.
        It then returns a list of Neuron instances with morphological data for each neuron.

        :param neurons: list of neurons metadata (as returned by one of the functions
                    used to fetch metadata)
        :param _name: used internally to save cached neurons with a different prefix when the
                class is used to download neurons for other APIs
        """
        if not isinstance(neurons, (list, tuple)):
            neurons = [neurons]

        to_return = []
        for neuron in neurons:
            if not isinstance(neuron, dict):
                raise ValueError()

            try:
                neuron["status"] == 500  # download went wrong
                continue
            except KeyError:
                pass

            filepath = self.build_filepath(neuron["neuron_id"])
            load_current_neuron = load_neurons

            if not os.path.isfile(filepath):
                # Download and write to file
                if self._version == "CNG version":
                    url = f"http://neuromorpho.org/dableFiles/{neuron['archive'].lower()}/CNG version/{neuron['neuron_name']}.CNG.swc"
                else:
                    url = f"http://neuromorpho.org/dableFiles/{neuron['archive'].lower()}/{self._version}/{neuron['neuron_name']}.swc"

                try:
                    req = request_no_ssl(url)
                    with open(filepath, "w") as f:
                        f.write(req.content.decode("utf-8"))
                except ValueError as exc:
                    logger.error(
                        "Could not fetch the neuron %s for the following reason: %s",
                        neuron["neuron_name"],
                        str(exc),
                    )
                    load_current_neuron = False

            if _name is None:
                _name = "neuromorpho_"

            to_return.append(
                Neuron(
                    filepath,
                    neuron_name=_name + str(neuron["neuron_id"]),
                    load_file=load_current_neuron,
                    **kwargs,
                )
            )

        return to_return



import logging
from collections import namedtuple

import pandas as pd
from bg_atlasapi import BrainGlobeAtlas
from retry import retry

from morphapi.api.neuromorphorg import NeuroMorpOrgAPI
from morphapi.morphology.morphology import Neuron
from morphapi.paths_manager import Paths
from morphapi.utils.data_io import flatten_list
from morphapi.utils.data_io import is_any_item_in_list
from morphapi.utils.webqueries import mouselight_base_url
from morphapi.utils.webqueries import post_mouselight

logger = logging.getLogger(__name__)


"""
    Collections of functions to query http://ml-neuronbrowser.janelia.org/ and get data about either the status of the API,
    the brain regions or the neurons available.
    Queries are sent by sending POST requests to http://ml-neuronbrowser.janelia.org/graphql
    with a string query.
"""

# ---------------------------------------------------------------------------- #
#                                  QUERY UTILS                                 #
# ---------------------------------------------------------------------------- #


def mouselight_api_info():
    """
    Get the number of cells available in the database
    """
    # Get info from the ML API
    url = mouselight_base_url + "graphql"

    query = """
            query {
                queryData {
                    totalCount
                    }
                }
            """
    res = post_mouselight(url, query=query)
    logger.info(
        "%s neurons on MouseLight database. ", res["queryData"]["totalCount"]
    )


def mouselight_get_brainregions():
    """
    Get metadata about the brain brain regions as they are known by Janelia's Mouse Light.
    IDs and Names sometimes differ from Allen's CCF.
    """

    # Download metadata about brain regions from the ML API
    url = mouselight_base_url + "graphql"
    # query =  "systemSettings {apiVersion apiRelease neuronCount\}}"
    query = """
            query {
                brainAreas{

                    acronym
                    name
                    id
                    atlasId
                    graphOrder
                    parentStructureId
                    structureIdPath
                }
            }
            """
    res = post_mouselight(url, query=query)["brainAreas"]

    # Clean up and turn into a dataframe
    keys = {k: [] for k in res[0].keys()}
    for r in res:
        for k in r.keys():
            keys[k].append(r[k])

    structures_data = pd.DataFrame.from_dict(keys)
    return structures_data


def mouselight_structures_identifiers():
    """
    When the data are downloaded as SWC, each node has a structure identifier ID to tell if it's soma, axon or dendrite.
    This function returns the ID number --> structure table.
    """

    # Download the identifiers used in ML neurons tracers
    url = mouselight_base_url + "graphql"
    # query =  "systemSettings {apiVersion apiRelease neuronCount\}}"
    query = """
            query {
                structureIdentifiers{
                    id
                    name
                    value
                }
            }
        """
    res = post_mouselight(url, query=query)["structureIdentifiers"]

    keys = {k: [] for k in res[0].keys()}
    for r in res:
        for k in r.keys():
            keys[k].append(r[k])

    structures_identifiers = pd.DataFrame.from_dict(keys)
    return structures_identifiers


def make_query(filterby=None, filter_regions=None, invert=False):
    """
    Constructs the strings used to submit graphql queries to the mouse light api

    :param filterby: str, soma, axon on dendrite. Search by neurite structure (Default value = None)
    :param filter_regions:  list, tuple. list of strings. Acronyms of brain regions to use for query (Default value = None)
    :param invert:  If true the inverse of the query is return (i.e. the neurons NOT in a brain region) (Default value = False)

    """
    searchneurons = """
                queryTime
                totalCount

                neurons{
                tag
                id
                idNumber
                idString

                brainArea{
                    id
                    acronym
                    name
                    safeName
                    atlasId
                    aliasList
                    structureIdPath
                }

                tracings{
                    soma{
                    x
                    y
                    z
                    radius
                    brainAreaIdCcfV30
                    sampleNumber
                    parentNumber

                    }

                id
                tracingStructure{
                    name
                    value
                    id
                }
                }
            }
    """

    if filterby is None or filterby == "soma":
        query = """
                    query {{
                        searchNeurons {{
                            {}
                        }}
                    }}
                    """.format(
            searchneurons
        )
    else:
        raise NotImplementedError("This feature is not available yet")
        # Get predicate type
        if filterby.lower() in ["axon", "axons", "end point", "branch point"]:
            predicateType = 1

        elif filterby.lower() in [
            "dendrite",
            "apical dendrite",
            "(basal) dendrite",
        ]:
            raise NotImplementedError
            filterby = "(basal) dendrite"
            predicateType = 2
        else:
            raise ValueError("invalid search by argument")

        # Get neuron structure id
        structures_identifiers = mouselight_structures_identifiers()
        structureid = str(
            structures_identifiers.loc[
                structures_identifiers.name == filterby
            ]["id"].values[0]
        )

        # Get brain regions ids
        brainregions = mouselight_get_brainregions()
        brainareaids = [
            str(brainregions.loc[brainregions.acronym == a]["id"].values[0])
            for a in filter_regions
        ]

        # Get inversion
        if invert:
            invert = "true"
        else:
            invert = "false"

        query = """
        query {{
            searchNeurons (
                context: {{
                scope: 6
                predicates: [{{
                    predicateType: {predicate}
                    tracingIdsOrDOIs: []
                    tracingIdsOrDOIsExactMatch: false
                    tracingStructureIds: []
                    amount: 0
                    nodeStructureIds: ['{structure}']
                    brainAreaIds: {brainarea}
                    invert: {invert}
                    composition: 1
                    }}]
                }}
            ) {{
                {base}
            }}
        }}
        """.format(
            predicate=predicateType,
            structure=str(structureid),
            brainarea=brainareaids,
            invert=invert,
            base=searchneurons,
        )

        query = query.replace("\\t", "").replace("'", '"')
    return query


@retry(tries=3, delay=1)
def fetch_atlas(atlas_name="allen_mouse_25um"):
    """Fetch a given atlas.

    See here for available atlases:
    https://docs.brainglobe.info/bg-atlasapi/introduction#atlases-available
    """
    return BrainGlobeAtlas(atlas_name)


# ---------------------------------------------------------------------------- #
#                                  MAIN CLASS                                  #
# ---------------------------------------------------------------------------- #


class MouseLightAPI(Paths):
    def __init__(self, base_dir=None, **kwargs):
        """
        Handles the download of neurons morphology data from the Mouse Light project

        :param base_dir: path to directory to use for saving data (default value None)
        :param kwargs: can be used to pass path to individual data folders. See morphapi/utils /paths_manager.py
        """
        Paths.__init__(self, base_dir=base_dir, **kwargs)

    def fetch_neurons_metadata(
        self, filterby=None, filter_regions=None, **kwargs
    ):
        """
        Download neurons metadata and data from the API. The downloaded metadata can be filtered to keep only
        the neurons whose soma is in a list of user selected brain regions.

        :param filterby: Accepted values: "soma". If it's "soma", neurons are kept only when their soma
                        is in the list of brain regions defined by filter_regions (Default value = None)
        :param filter_regions: List of brain regions acronyms. If filtering neurons, these specify the filter criteria. (Default value = None)
        :param **kwargs:

        """
        # Download all metadata
        logger.debug("Querying MouseLight API...")
        url = mouselight_base_url + "graphql"
        query = make_query(
            filterby=filterby, filter_regions=filter_regions, **kwargs
        )

        res = post_mouselight(url, query=query)["searchNeurons"]
        logger.info(
            "Fetched metadata for %s neurons in %ss",
            res["totalCount"],
            round(res["queryTime"] / 1000, 2),
        )

        # Process neurons to clean up the results and make them easier to handle
        neurons = res["neurons"]
        node = namedtuple("node", "x y z r area_acronym sample_n parent_n")
        tracing_structure = namedtuple(
            "tracing_structure", "id name value named_id"
        )

        cleaned_neurons = []  # <- output is stored here
        for neuron in neurons:
            if neuron["brainArea"] is not None:
                brainArea_acronym = neuron["brainArea"]["acronym"]
                brainArea_id = neuron["brainArea"]["id"]
                brainArea_name = neuron["brainArea"]["name"]
                brainArea_safename = neuron["brainArea"]["safeName"]
                brainArea_atlasId = neuron["brainArea"]["atlasId"]
                brainArea_structureIdPath = neuron["brainArea"][
                    "structureIdPath"
                ]
            else:
                brainArea_acronym = None
                brainArea_id = None
                brainArea_name = None
                brainArea_safename = None
                brainArea_atlasId = None
                brainArea_structureIdPath = None

            if len(neuron["tracings"]) > 1:
                dendrite = tracing_structure(
                    neuron["tracings"][1]["id"],
                    neuron["tracings"][1]["tracingStructure"]["name"],
                    neuron["tracings"][1]["tracingStructure"]["value"],
                    neuron["tracings"][1]["tracingStructure"]["id"],
                )
            else:
                dendrite = None

            clean_neuron = dict(
                brainArea_acronym=brainArea_acronym,
                brainArea_id=brainArea_id,
                brainArea_name=brainArea_name,
                brainArea_safename=brainArea_safename,
                brainArea_atlasId=brainArea_atlasId,
                brainArea_structureIdPath=brainArea_structureIdPath,
                id=neuron["id"],
                idNumber=neuron["idNumber"],
                idString=neuron["idString"],
                tag=neuron["tag"],
                soma=node(
                    neuron["tracings"][0]["soma"]["x"],
                    neuron["tracings"][0]["soma"]["y"],
                    neuron["tracings"][0]["soma"]["z"],
                    neuron["tracings"][0]["soma"]["radius"],
                    brainArea_name,
                    neuron["tracings"][0]["soma"]["sampleNumber"],
                    neuron["tracings"][0]["soma"]["parentNumber"],
                ),
                axon=tracing_structure(
                    neuron["tracings"][0]["id"],
                    neuron["tracings"][0]["tracingStructure"]["name"],
                    neuron["tracings"][0]["tracingStructure"]["value"],
                    neuron["tracings"][0]["tracingStructure"]["id"],
                ),
                dendrite=dendrite,
            )
            cleaned_neurons.append(clean_neuron)

        if filter_regions is not None:
            cleaned_neurons = self.filter_neurons_metadata(
                cleaned_neurons,
                filterby=filterby,
                filter_regions=filter_regions,
            )

        return cleaned_neurons

    @staticmethod
    def fetch_default_atlas():
        """Fetch the allen mouse 25nm atlas."""
        return fetch_atlas("allen_mouse_25um")

    def filter_neurons_metadata(
        self,
        neurons_metadata,
        filterby="soma",
        filter_regions=None,
        atlas=None,
    ):
        """
        Filter metadata to keep only the neurons whose soma is in a given list of brain regions.

        :param filterby: Accepted values: "soma". If it's "soma", neurons are kept only when their
                         soma is in the list of brain regions defined by filter_regions (Default
                         value = None)
        :param filter_regions: List of brain regions acronyms. If filtering neurons, these specify
                               the filter criteria. (Default value = None)
        :param atlas: A `bg_atlasapi.BrainGlobeAtlas` object. If not provided, load the
                      default atlas.
        """

        # Filter neurons to keep only those matching the search criteria
        if filterby is None:
            logger.info(
                "Returning all neurons because 'filterby' is set to None",
            )
            return neurons_metadata

        if filter_regions is None:
            raise ValueError(
                "If filtering neuron by region, you need to pass a list of filter regions to use"
            )

        # get brain globe atlas
        if atlas is None:
            atlas = self.fetch_default_atlas()

        # Filter by soma
        if filterby == "soma":
            filtered_neurons_metadata = []
            for neuron in neurons_metadata:
                if neuron["brainArea_acronym"] is None:
                    continue

                # get ancestors of neuron's regions
                try:
                    neuron_region_ancestors = atlas.get_structure_ancestors(
                        neuron["brainArea_acronym"]
                    )
                    neuron_region_ancestors.append(neuron["brainArea_acronym"])
                except KeyError:
                    # ignore if region is not found
                    continue

                # If any of the ancestors or itself are in the allowed regions, keep neuron.
                if is_any_item_in_list(
                    filter_regions, neuron_region_ancestors
                ):
                    filtered_neurons_metadata.append(neuron)

            neurons = filtered_neurons_metadata
        else:
            neurons = neurons_metadata

        logger.info(
            "Selected %s neurons out of %s",
            len(neurons),
            len(neurons_metadata),
        )

        return neurons

    def download_neurons(self, neurons_metadata, load_neurons=True, **kwargs):
        """
        Given a list of neurons metadata from self.fetch_neurons_metadata
        this funcition downloads the morphological data.
        The data are actually downloaded from neuromorpho.org

        :param neurons_metadata: list with metadata for neurons to download
        :returns: list of Neuron instances

        """
        if not isinstance(neurons_metadata, (list, tuple)):
            neurons_metadata = [neurons_metadata]

        nmapi = NeuroMorpOrgAPI()
        nmapi._version = "Source-Version"
        nmapi.neuromorphorg_cache = self.mouselight_cache

        neurons = []

        for neuron in neurons_metadata:
            try:
                nrn = nmapi.get_neuron_by_name(neuron["idString"])
            except ValueError as exc:
                logger.error(
                    "Could not fetch the neuron %s for the following reason: %s",
                    neuron["idString"],
                    str(exc),
                )
                neurons.append(
                    Neuron(
                        nmapi.build_filepath(neuron["idString"]),
                        neuron_name="mouselight_" + str(neuron["idString"]),
                        invert_dims=True,
                        load_file=False,
                    )
                )
                continue

            downloaded = nmapi.download_neurons(
                nrn,
                _name="mouselight_",
                invert_dims=True,
                load_neurons=load_neurons,
            )

            neurons.append(downloaded)

        return flatten_list(neurons)
